// DPCplus ARM code demo
// Last Update:  May 30, 2010
//              Darrell Spice, Jr.

// To compile, this you'll need to install an ARM compiler.
// for OS X use:
// http://www.gnuarm.com/files.html
//
// this demo was written using the OS X GCC-3.3 toolchain
// http://www.gnuarm.com/binutils-2.14_gcc-3.3.2-c-c++_newlib-1.12.0_gdb-6.0.pkg.tgz
//
//
// for Linux/Windows use:
// http://www.codesourcery.com/sgpp/lite/arm/portal/release1033
//
// You will need to update the "Makefile" if you're using Linux/Windows.  Find the toolchain lines
// and move the # from the aebi toolchain to the elf version.
//  #TOOLCHAIN=arm-elf
//  TOOLCHAIN=arm-none-eabi



// included for functions my_memset, my_memcpy
//#include <string.h>



// src/custom.h defines the following:
//
// queue[]     - points to the 4K Display Data Bank
//             - treat as RAM
//             - Any data passed to/from the ARM and 6507 must be done via queue[]
//
// flashdata[] - points to the 24K that comprises the six 4K banks.
//             - treat as ROM
//						 - yes, you can access data stored in the "6507 banks", see gDigitsPtr for an example
#include "src/custom.h"



// types of variables, storage used and range
// const              = constant, which is compiled into the ROM section and cannot be changed during run time.
// char               =  8 bit, 1 byte  per value.  Range is           -128 to           127
// unsigned char      =  8 bit, 1 byte  per value.  Range is              0 to           255
// short int          = 16 bit, 2 bytes per value.  Range is        -32,768 to        32,767
// unsigned short int = 16 bit, 2 bytes per value.  Range is        -32,768 to        32,767
// int                = 32 bit, 4 bytes per value.  Range is -2,147,483,648 to 2,147,483,647
// unsigned int       = 32 bit, 4 bytes per value.  Range is              0 to 4,294,967,295
// long int           = same as int
// float              = 32 bit, 4 bytes.  Not Supported#
// double             = 64 bit, 8 bytes.  Not Supported#
// *                  = 32 bit, 4 bytes, pointer (ie: char*, int*)
//
// # - at least not supported with the compiler under OS X.  It might be supported by the Linux/Windows ARM
//     compiler - but support would be implemented via software as the ARM in the Harmony Cartridge does not
//     have an FPU (floating point unit).  As such, performance will most likely be as good as using
//     integer math based routines.
//
// NOTE : Only 448 bytes are allocated for use by variables (the rest of the 512 byte section is used 
//        as the stack).  If you use too many variables, you'll get a compile time error:  "region ram is full"
//        You can use RAM in the Display Data to store values, as shown by the following defines:
// globals stored in queue
#define gBitmapWidth		queue[0x0fff]
#define gBitmapHeight		queue[0x0ffe]



// SIN and COS tables are defined using 0-255 instead of 0-360 (ie: 64 = 90 degrees, 128 = 180 degrees, etc).
// this lets us use an unsigned char (1 byte) to hold rotation amounts instead of 2 bytes
// these values are also multipled by 2^12 as we're using integer math
//
// these are defined in tables.c and were creating using the OpenOffice spreadsheet "Sin Cos Tables.ods"
//
// if adding or removing files to the C project, be sure to update the CUSTOMOBJS line in the "Makefile".
// 
// as noted above, "const" means these values are stored in ROM and cannot be changed
extern const short int sin_table[256];
extern const short int cos_table[256];



// global variables are those found outside of any function.  They can be used by any function.
// The Harmony Cartridge preserves these values for you between ARM code calls.
int gBitmapAddress;
int gDisplayPoints[8][2];

unsigned char gXrotation;
unsigned char gYrotation;
unsigned char gZrotation;
unsigned char gScale;


unsigned char* gBitmapPtr;
unsigned char* gDigitsPtr;

const short int cube_local_vertices[8][4] = 
{	// vertices = points
  {+20, +20, +20, 1},	// 0
	{-20, +20, +20, 1},	// 1
	{-20, +20, -20, 1},	// 2
	{+20, +20, -20, 1},	// 3
	{+20, -20, +20, 1},	// 4
	{-20, -20, +20, 1},	// 5
	{-20, -20, -20, 1},	// 6
	{+20, -20, -20, 1}	// 7
};

const short int cube_vectors[12][2] =
{	// vectors = lines joining points
	{0, 1},
	{0, 3},
	{0, 4},
	{1, 5},
	{1, 2},
	{2, 3},
	{2, 6},
	{3, 7},
	{4, 5},
	{4, 7},
	{5, 6},
	{6, 7}	
};

short int matrix1[8][4];
short int matrix2[8][4];
short int matrix[4][4];



// Function Prototypes are required for functions that are defined AFTER they are used
// ie: main() calls Initialize(), which is defined later in the source file, so Initialize() must be prototyped
int Initialize();
int OverScan();
int VerticalBlank();
inline void SetPoint(int x, int y);
void ClearPoint(int x, int y);
void DrawLine(int x0, int y0, int x1, int y1);
void matrix_multiply(short int initial[][4], short int transform[][4], short int result[][4], short int rows);


// note - use VOID for functions that do not return a value
// as they will take up less space in the compiled code
void my_memcpy(unsigned char* destination, unsigned char* source, int count)
{
	int i;
	for(i=0;i<count;i++)
		destination[i] = source[i];
}

void my_memset(unsigned char* destination, int fill_value, int count)
{
	int i;
	for (i=0;i<count;i++)
		destination[i]=fill_value;
}



// main() is what gets called when you store $FF into DPC+ register CALLFUNCTION in your 6507 code.
int main()
{
	// We're using the first byte of Display Data to tell the ARM code which subroutine we wish to run
	// in the 6507 code this is defined as ARMsub
	switch (queue[0x0])	
	{
		case 0:
			Initialize();
			break;
		case 1:
			OverScan();
			break;
		case 2:
			VerticalBlank();
			break;
	}
	return 0;
}

 

// The Initialize() function is used to prep the ARM code with the size and location of the Bitmap Data.
int Initialize()
{
	gBitmapWidth		= queue[0x01];
	gBitmapHeight		= queue[0x02];
	gBitmapAddress	= queue[0x03] + queue[0x04] * 256;
	
	gBitmapPtr = (unsigned char *)queue + gBitmapAddress;
	gDigitsPtr = (unsigned char *)flashdata + 0x5F00;
	
  gXrotation = gYrotation = gZrotation = 0;
	gScale = 0x20;

	return 0;
}



// The OverScan function is called at the start of every OverScan.  It:
//   - clears the bitmap,
//   - processes joystick readings
//   - calculates 3D graphic data
int OverScan()
{
	int i;
	unsigned char	SWCHA		= queue[0x01];
//	unsigned char SWCHB		= queue[0x02];	// these aren't used so are commented out
//	unsigned char CXM0P		= queue[0x03];  // to suppress "unused variable" compiler warnings
//	unsigned char CXM1P		= queue[0x04];  // I've left them here as a template for your projects
//	unsigned char CXP0FB	= queue[0x05];
//	unsigned char CXP1FB	= queue[0x06];
//	unsigned char CXM0FB	= queue[0x07];
//	unsigned char CXM1FB	= queue[0x08];
//	unsigned char CXBLPF	= queue[0x09];
//	unsigned char CXPPMM	= queue[0x0a];
	unsigned char INPT4		= queue[0x0b];
//	unsigned char INPT5		= queue[0x0c];
	
	
	// my_memset fills memory, in this case we're using it to clear the bitamp by filling it with 0
		my_memset(gBitmapPtr, 0, gBitmapWidth * gBitmapHeight/8);
	
	// process joystick input
	if (INPT4 & 0x80) 
	{	// fire button not pressed
		if (!(SWCHA & 0x80)) // joystick right
			gXrotation++;
		if (!(SWCHA & 0x40)) // joystick left
			gXrotation--;
		if (!(SWCHA & 0x20)) // joystick down
			gYrotation--;
		if (!(SWCHA & 0x10)) // joystick up
			gYrotation++;
	}
	else 
	{	// fire button pressed
		if (!(SWCHA & 0x80)) // joystick right
			gZrotation++;
		if (!(SWCHA & 0x40)) // joystick left
			gZrotation--;
		if (!(SWCHA & 0x20)) // joystick down
			gScale--;
		if (!(SWCHA & 0x10)) // joystick up
			gScale++;
	}
	
	gScale &= 0x3f;	// limit gScale to values from 0-63
	
	// update 3D object based on X rotation
	matrix[0][0] = 1;			matrix[0][1] = 0;													matrix[0][2] = 0;													matrix[0][3] = 0;
	matrix[1][0] = 0;			matrix[1][1] = cos_table[gXrotation];			matrix[1][2] = sin_table[gXrotation];			matrix[1][3] = 0;
	matrix[2][0] = 0;			matrix[2][1] = -sin_table[gXrotation];		matrix[2][2] = cos_table[gXrotation];			matrix[2][3] = 0;
	matrix[3][0] = 0;			matrix[3][1] = 0;													matrix[3][2] = 0;													matrix[3][3] = 1;
	
	matrix_multiply(cube_local_vertices, matrix, matrix2, 8);
	
	// update 3D object based on Y rotation
	matrix[0][0] = cos_table[gYrotation];			matrix[0][1] = 0;			matrix[0][2] = -sin_table[gYrotation];		matrix[0][3] = 0;
	matrix[1][0] = 0;													matrix[1][1] = 1;			matrix[1][2] = 0;													matrix[1][3] = 0;
	matrix[2][0] = sin_table[gYrotation];			matrix[2][1] = 0;			matrix[2][2] = cos_table[gYrotation];			matrix[2][3] = 0;
	matrix[3][0] = 0;													matrix[3][1] = 0;			matrix[3][2] = 0;													matrix[3][3] = 1;

	matrix_multiply(matrix2, matrix, matrix1, 8);
	
	// update 3D object based on Z rotation
	matrix[0][0] = cos_table[gZrotation];			matrix[0][1] = sin_table[gZrotation];		matrix[0][2] = 0;				matrix[0][3] = 0;
	matrix[1][0] = -sin_table[gZrotation];		matrix[1][1] = cos_table[gZrotation];		matrix[1][2] = 0;				matrix[1][3] = 0;
	matrix[2][0] = 0;													matrix[2][1] = 0;												matrix[2][2] = 1;				matrix[2][3] = 0;
	matrix[3][0] = 0;													matrix[3][1] = 0;												matrix[3][2] = 0;				matrix[3][3] = 1;

	matrix_multiply(matrix1, matrix, matrix2, 8);
	
	// shift object to middle of bitmap
	matrix[0][0] = 1;									matrix[0][1] = 0;									matrix[0][2] = 0;			matrix[0][3] = 0;
	matrix[1][0] = 0;									matrix[1][1] = 1;									matrix[1][2] = 0;			matrix[1][3] = 0;
	matrix[2][0] = 0;									matrix[2][1] = 0;									matrix[2][2] = 1;			matrix[2][3] = 0;
	matrix[3][0] = gBitmapWidth/2;		matrix[3][1] = gBitmapHeight/2;		matrix[3][2] = 100;		matrix[3][3] = 1;

	matrix_multiply(matrix2, matrix, matrix1, 8);
	
	// 3D to 2D projection
	for (i = 0; i< 8; i++)
	{
		gDisplayPoints[i][0] = gScale * matrix1[i][0] / matrix1[i][2];	// screen x
		gDisplayPoints[i][1] = gScale * matrix1[i][1] / matrix1[i][2];	// screen y
	}
	
	// show X rotation value
	my_memcpy(gBitmapPtr,											gDigitsPtr + ((gXrotation & 0xF0) >> 1), 8);
	my_memcpy(gBitmapPtr + gBitmapHeight * 1,	gDigitsPtr + ((gXrotation & 0x0F) << 3), 8);

	// show Y rotation value
	my_memcpy(gBitmapPtr + gBitmapHeight * 3,	gDigitsPtr + ((gYrotation & 0xF0) >> 1), 8);
	my_memcpy(gBitmapPtr + gBitmapHeight * 4,	gDigitsPtr + ((gYrotation & 0x0F) << 3), 8);
	
	// show Z rotation value
	my_memcpy(gBitmapPtr + gBitmapHeight * 6,	gDigitsPtr + ((gZrotation & 0xF0) >> 1), 8);
	my_memcpy(gBitmapPtr + gBitmapHeight * 7,	gDigitsPtr + ((gZrotation & 0x0F) << 3), 8);
	
	// show scale value
	my_memcpy(gBitmapPtr + gBitmapHeight * 9,	gDigitsPtr + ((gScale & 0xF0) >> 1), 8);
	my_memcpy(gBitmapPtr + gBitmapHeight * 10,	gDigitsPtr + ((gScale & 0x0F) << 3), 8);
	
	return 0;
}



// The VerticalBlank function is called at the start of every OverScan.
// it draws the 3D object.
int VerticalBlank()
{
	int i;
	
	for (i = 0; i < 12; i++)
		DrawLine(gDisplayPoints[cube_vectors[i][0]][0], gDisplayPoints[cube_vectors[i][0]][1],
						 gDisplayPoints[cube_vectors[i][1]][0], gDisplayPoints[cube_vectors[i][1]][1]);
	
	return 0;
}



// multiply 2 matrixes
void matrix_multiply(short int initial[][4], short int transform[][4], short int result[][4], short int rows)
{
	int i, j, k, sum;
	
	for(i=0;i< rows;i++)
		for(j=0;j< 4;j++)
		{
			sum = 0;
			for(k=0;k< 4;k++)
				sum +=  ((-256 < transform[k][j]) && (transform[k][j] < 256)) ? (initial[i][k] * transform[k][j]) : ((int)((initial[i][k] * transform[k][j])) >> 14); 
			result[i][j] = sum;
		}
}




// Fast version of DrawLine routine that adjusts a pointer and mask instead of calculating every X-Y address.
// it's complicated to follow so check out the ORIGINAL DrawLine ROUTINE at the end of this file

// macros from supercat
#define move_left() ( (x0--, cursor_mask <<= 1) || (cursor_mask = 1, cursor_addr -= gBitmapHeight) )
#define move_right() ( (x0++, cursor_mask >>= 1) || (cursor_mask = 128, cursor_addr += gBitmapHeight) )
#define move_up() (y0--, cursor_addr--)
#define move_down() (y0++, cursor_addr++)
#define pset_raw() (*cursor_addr |= cursor_mask)
#define pclr_raw() (*cursor_addr &= ~cursor_mask)
#define pset() ((x0 >= 0) && (x0 < gBitmapWidth) && (y0 >> 0) && (y0 < gBitmapHeight) && pset_raw())
#define pclr() ((x0 >= 0) && (x0 < gBitmapWidth) && (y0 >> 0) && (y0 < gBitmapHeight) && pclr_raw())

void DrawLine(int x0, int y0, int x1, int y1)
{
	int dy = y1 - y0;
	int dx = x1 - x0;
	int stepy, stepx, fraction;
	
	unsigned char *cursor_addr;
	unsigned char cursor_mask;
	
	if (dy < 0) {dy = -dy; stepy = -1;} else {stepy = 1;}
	if (dx < 0) {dx = -dx; stepx = -1;} else {stepx = 1;}
	dy <<= 1;
	dx <<= 1;
	
	cursor_addr = gBitmapPtr + (x0>>3)*gBitmapHeight + y0;
	cursor_mask = 1 << (7-(x0 & 0x07));
	pset();
	
	if (dx > dy)
	{
		fraction = dy - (dx >> 1);                         // same as 2*dy - dx
		if ((stepy < 0) && (stepx < 0))
		{
			while (x0 != x1)
			{
				if (fraction >= 0)
				{
					move_up();
					fraction -= dx;                                // same as fraction -= 2*dx
				}
				move_left();
				fraction += dy;                                    // same as fraction -= 2*dy
				pset();
			}			
		}
		else if ((stepy > 0) && (stepx < 0))
		{
			while (x0 != x1)
			{
				if (fraction >= 0)
				{
					move_down();
					fraction -= dx;                                // same as fraction -= 2*dx
				}
				move_left();
				fraction += dy;                                    // same as fraction -= 2*dy
				pset();
			}
		}
		else if ((stepy < 0) && (stepx > 0))
		{
			while (x0 != x1)
			{
				if (fraction >= 0)
				{
					move_up();
					fraction -= dx;                                // same as fraction -= 2*dx
				}
				move_right();
				fraction += dy;                                    // same as fraction -= 2*dy
				pset();
			}
		}
		else	// stepy > 0 && stepx > 0
		{
			while (x0 != x1)
			{
				if (fraction >= 0)
				{
					move_down();
					fraction -= dx;                                // same as fraction -= 2*dx
				}
				move_right();
				fraction += dy;                                    // same as fraction -= 2*dy
				pset();
			}
		}
	} 
	else
	{
		fraction = dx - (dy >> 1);
		if ((stepy < 0) && (stepx < 0))
		{
			while (y0 != y1)
			{
				if (fraction >= 0)
				{
					move_left();
					fraction -= dy;
				}
				move_up();
				fraction += dx;
				pset();
			}
		}
		else if ((stepy > 0) && (stepx < 0))
		{
			while (y0 != y1)
			{
				if (fraction >= 0)
				{
					move_left();
					fraction -= dy;
				}
				move_down();
				fraction += dx;
				pset();
			}
		}
		else if ((stepy < 0) && (stepx > 0))
		{
			while (y0 != y1)
			{
				if (fraction >= 0)
				{
					move_right();
					fraction -= dy;
				}
				move_up();
				fraction += dx;
				pset();
			}
		}
		else	// stepy > 0 && stepx > 0
		{
			while (y0 != y1)
			{
				if (fraction >= 0)
				{
					move_right();
					fraction -= dy;
				}
				move_down();
				fraction += dx;
				pset();
			}
		}
	}
}



/* THE ORIGINAL DrawLine ROUTINE with companion SetPoint function.
 
inline void SetPoint(int x, int y)
{
	if ((x >=0 && x < gBitmapWidth) && (y >= 0 && y < gBitmapHeight))
		gBitmapPtr[(x>>3)*gBitmapHeight + y] |= 1 << (7-(x & 0x07));	// 11-14 w/ bitmap[]

}

void DrawLine(int x0, int y0, int x1, int y1)
{
	// lineBresenham();
	// http://www.cs.unc.edu/~mcmillan/comp136/Lecture6/Lines.html
	int dy = y1 - y0;
	int dx = x1 - x0;
	int stepy, stepx, fraction;
	
	if (dy < 0) {dy = -dy; stepy = -1;} else {stepy = 1;}
	if (dx < 0) {dx = -dx; stepx = -1;} else {stepx = 1;}
	dy <<= 1;
	dx <<= 1;
	
	SetPoint(x0,y0);

	
	if (dx > dy)
	{
		fraction = dy - (dx >> 1);                         // same as 2*dy - dx
		while (x0 != x1)
		{
			if (fraction >= 0)
			{
				y0 += stepy;
				fraction -= dx;                                // same as fraction -= 2*dx
			}
			x0 += stepx;
			fraction += dy;                                    // same as fraction -= 2*dy
			SetPoint(x0, y0);
		}
	} 
	else
	{
		fraction = dx - (dy >> 1);
		while (y0 != y1)
		{
			if (fraction >= 0)
			{
				x0 += stepx;
				fraction -= dy;
			}
			y0 += stepy;
			fraction += dx;
			SetPoint(x0, y0);
		}
	}

}
 
*/


